import numpy as np 
import matplotlib.pyplot as plt
from warnings import warn

from ._tree import recurse_leaf_dfs, _compute_stability, _label_of_stability, _get_selected_clusters
from .cluster import Clustering

#COPIED FROM HDBSCAN LIB
CB_LEFT = 0
CB_RIGHT = 1
CB_BOTTOM = 2
CB_TOP = 3


def _recurse_leaf_dfs(cluster_tree, current_node):
    # COPIED FROM HDBSCAN LIB
    children = cluster_tree[cluster_tree['parent'] == current_node]['child']
    if len(children) == 0:
        return [current_node,]
    else:
        return sum([recurse_leaf_dfs(cluster_tree, child) for child in children], [])



def _get_leaves(condensed_tree):
    #COPIED FROM HDBSCAN LIB
    cluster_tree = condensed_tree[condensed_tree['child_size'] > 1]
    if cluster_tree.shape[0] == 0:
        # Return the only cluster, the root
        return [condensed_tree['parent'].min()]

    root = cluster_tree['parent'].min()
    return _recurse_leaf_dfs(cluster_tree, root)


class CondensedTree(object):
    """ Implementation of the result of the runt pruning on a linkage tree

    Attributes
    ----------
        condensed_tree : np.ndarray that respresents the result of the runt pruning procedure on the linkage tree

    The representation of condensed tree is a numpy array of edges with the form (p,c,l,s) with
        - p parent
        - c children
        - v lambda value 
        - s size of the child
    """

    def __init__(self, raw_tree = None, path = None):
        if path != None : 
            if path[-4:] != ".npy":
                raise AttributeError("The linkage tree is suppose to be generated from a csv file.")
            else:
                self._raw_tree = np.load(path)
        else :
            self._raw_tree = raw_tree
        self._clusters_stability = None

    @property
    def raw_tree(self):
        """ .. :no-docs:
        """
        if self._raw_tree is None :
            raise AttributeError("Raw tree has not been initialised")
        else :
            return self._raw_tree

    @property
    def clusters_stability(self):
        """ .. :no-docs:
        """
        if self._clusters_stability is None : 
            raise AttributeError("Clusters' stabliity was not computed yet. Run self.compute_stability.")
        else : 
            return self._clusters_stability

    
    def get_plot_data(self,
                      leaf_separation=1,
                      log_size=False,
                      max_rectangle_per_icicle=20):

        #COPIED FROM HDBSCAN LIB
        """Generates data for use in plotting the 'icicle plot' or dendrogram
        plot of the condensed tree generated by HDBSCAN.

        Parameters
        ----------
        leaf_separation : float, optional
                          How far apart to space the final leaves of the
                          dendrogram. (default 1)

        log_size : boolean, optional
                   Use log scale for the 'size' of clusters (i.e. number of
                   points in the cluster at a given lambda value).
                   (default False)

        max_rectangles_per_icicle : int, optional
            To simplify the plot this method will only emit
            ``max_rectangles_per_icicle`` bars per branch of the dendrogram.
            This ensures that we don't suffer from massive overplotting in
            cases with a lot of data points.

        Returns
        -------
        plot_data : dict
                    Data associated to bars in a bar plot:
                        `bar_centers` x coordinate centers for bars
                        `bar_tops` heights of bars in lambda scale
                        `bar_bottoms` y coordinate of bottoms of bars
                        `bar_widths` widths of the bars (in x coord scale)
                        `bar_bounds` a 4-tuple of [left, right, bottom, top]
                                     giving the bounds on a full set of
                                     cluster bars
                    Data associates with cluster splits:
                        `line_xs` x coordinates for horizontal dendrogram lines
                        `line_ys` y coordinates for horizontal dendrogram lines
        """
        leaves = _get_leaves(self._raw_tree)
        last_leaf = self._raw_tree['parent'].max()
        root = self._raw_tree['parent'].min()

        # We want to get the x and y coordinates for the start of each cluster
        # Initialize the leaves, since we know where they go, the iterate
        # through everything from the leaves back, setting coords as we go
        if isinstance(leaves, np.int64):
            cluster_x_coords = {leaves: leaf_separation}
        else:
            cluster_x_coords = dict(zip(leaves, [leaf_separation * x
                                                 for x in range(len(leaves))]))
        cluster_y_coords = {root: 0.0}

        for cluster in range(last_leaf, root - 1, -1):
            split = self._raw_tree[['child', 'lamb_val']]
            split = split[(self._raw_tree['parent'] == cluster) &
                          (self._raw_tree['child_size'] > 1)]
            if len(split['child']) > 1:
                left_child, right_child = split['child']
                cluster_x_coords[cluster] = np.mean([cluster_x_coords[left_child],
                                                     cluster_x_coords[right_child]])
                cluster_y_coords[left_child] = split['lamb_val'][0]
                cluster_y_coords[right_child] = split['lamb_val'][1]

        # We use bars to plot the 'icicles', so we need to generate centers, tops,
        # bottoms and widths for each rectangle. We can go through each cluster
        # and do this for each in turn.
        bar_centers = []
        bar_tops = []
        bar_bottoms = []
        bar_widths = []

        cluster_bounds = {}

        scaling = np.sum(self._raw_tree[self._raw_tree['parent'] == root]['child_size'])

        if log_size:
            scaling = np.log(scaling)

        for c in range(last_leaf, root - 1, -1):

            cluster_bounds[c] = [0, 0, 0, 0]

            c_children = self._raw_tree[self._raw_tree['parent'] == c]
            current_size = np.sum(c_children['child_size'])
            current_lambda = cluster_y_coords[c]
            cluster_max_size = current_size
            cluster_max_lambda = c_children['lamb_val'].max()
            cluster_min_size = np.sum(
                c_children[c_children['lamb_val'] ==
                           cluster_max_lambda]['child_size'])

            if log_size:
                current_size = np.log(current_size)
                cluster_max_size = np.log(cluster_max_size)
                cluster_min_size = np.log(cluster_min_size)

            total_size_change = float(cluster_max_size - cluster_min_size)
            step_size_change = total_size_change / max_rectangle_per_icicle

            cluster_bounds[c][CB_LEFT] = cluster_x_coords[c] * scaling - (current_size / 2.0)
            cluster_bounds[c][CB_RIGHT] = cluster_x_coords[c] * scaling + (current_size / 2.0)
            cluster_bounds[c][CB_BOTTOM] = cluster_y_coords[c]
            cluster_bounds[c][CB_TOP] = np.max(c_children['lamb_val'])
            

            last_step_size = current_size
            last_step_lambda = current_lambda

            for i in np.argsort(c_children['lamb_val']):
                row = c_children[i]
                if row['lamb_val'] != current_lambda and \
                        (last_step_size - current_size > step_size_change
                        or row['lamb_val'] == cluster_max_lambda):
                    bar_centers.append(cluster_x_coords[c] * scaling)
                    bar_tops.append(row['lamb_val'] - last_step_lambda)
                    bar_bottoms.append(last_step_lambda)
                    bar_widths.append(last_step_size)
                    last_step_size = current_size
                    last_step_lambda = current_lambda
                if log_size:
                    exp_size = np.exp(current_size) - row['child_size']
                    # Ensure we don't try to take log of zero
                    if exp_size > 0.01:
                        current_size = np.log(np.exp(current_size) - row['child_size'])
                    else:
                        current_size = 0.0
                else:
                    current_size -= row['child_size']
                current_lambda = row['lamb_val']

        # Finally we need the horizontal lines that occur at cluster splits.
        line_xs = []
        line_ys = []

        for row in self._raw_tree[self._raw_tree['child_size'] > 1]:
            parent = row['parent']
            child = row['child']
            child_size = row['child_size']
            if log_size:
                child_size = np.log(child_size)
            sign = np.sign(cluster_x_coords[child] - cluster_x_coords[parent])
            line_xs.append([
                cluster_x_coords[parent] * scaling,
                cluster_x_coords[child] * scaling + sign * (child_size / 2.0)
            ])
            line_ys.append([
                cluster_y_coords[child],
                cluster_y_coords[child]
            ])

        return {
            'bar_centers': bar_centers,
            'bar_tops': bar_tops,
            'bar_bottoms': bar_bottoms,
            'bar_widths': bar_widths,
            'line_xs': line_xs,
            'line_ys': line_ys,
            'cluster_bounds': cluster_bounds
        }



    def plot(self, leaf_separation=1, cmap='plasma', select_clusters=False,
                label_clusters=False, selection_palette=None,
                axis=None, colorbar=True, log_size=False,
                max_rectangles_per_icicle=20):
            
            # COPIED FROM HDBSCAN LIB
            
            """Use matplotlib to plot an 'icicle plot' dendrogram of the condensed tree.

            Effectively this is a dendrogram where the width of each cluster bar is
            equal to the number of points (or log of the number of points) in the cluster
            at the given lambda value. Thus bars narrow as points progressively drop
            out of clusters. The make the effect more apparent the bars are also colored
            according the the number of points (or log of the number of points).

            Parameters
            ----------
            leaf_separation : float, optional (default 1)
                            How far apart to space the final leaves of the
                            dendrogram.

            cmap : string or matplotlib colormap, optional (default plasma)
                The matplotlib colormap to use to color the cluster bars.


            select_clusters : boolean, optional (default False)
                            Whether to draw ovals highlighting which cluster
                            bar represent the clusters that were selected by
                            HDBSCAN as the final clusters.

            label_clusters : boolean, optional (default False)
                            If select_clusters is True then this determines
                            whether to draw text labels on the clusters.

            selection_palette : list of colors, optional (default None)
                                If not None, and at least as long as
                                the number of clusters, draw ovals
                                in colors iterating through this palette.
                                This can aid in cluster identification
                                when plotting.

            axis : matplotlib axis or None, optional (default None)
                The matplotlib axis to render to. If None then a new axis
                will be generated. The rendered axis will be returned.


            colorbar : boolean, optional (default True)
                    Whether to draw a matplotlib colorbar displaying the range
                    of cluster sizes as per the colormap.

            log_size : boolean, optional (default False)
                    Use log scale for the 'size' of clusters (i.e. number of
                    points in the cluster at a given lambda value).


            max_rectangles_per_icicle : int, optional (default 20)
                To simplify the plot this method will only emit
                ``max_rectangles_per_icicle`` bars per branch of the dendrogram.
                This ensures that we don't suffer from massive overplotting in
                cases with a lot of data points.

            Returns
            -------
            axis : matplotlib axis
                The axis on which the 'icicle plot' has been rendered.
            """
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                raise ImportError(
                    'You must install the matplotlib library to plot the condensed tree.'
                    'Use get_plot_data to calculate the relevant data without plotting.')

            plot_data = self.get_plot_data(leaf_separation=leaf_separation,
                                        log_size=log_size,
                                        max_rectangle_per_icicle=max_rectangles_per_icicle)

            if cmap != 'none':
                sm = plt.cm.ScalarMappable(cmap=cmap,
                                        norm=plt.Normalize(0, max(plot_data['bar_widths'])))
                sm.set_array(plot_data['bar_widths'])
                bar_colors = [sm.to_rgba(x) for x in plot_data['bar_widths']]
            else:
                bar_colors = 'black'

            if axis is None:
                axis = plt.gca()

            axis.bar(
                plot_data['bar_centers'],
                plot_data['bar_tops'],
                bottom=plot_data['bar_bottoms'],
                width=plot_data['bar_widths'],
                color=bar_colors,
                align='center',
                linewidth=0
            )

            drawlines = []
            for xs, ys in zip(plot_data['line_xs'], plot_data['line_ys']):
                drawlines.append(xs)
                drawlines.append(ys)
            axis.plot(*drawlines, color='black', linewidth=1)
            # for xs, ys in zip(plot_data['line_xs'], plot_data['line_ys']):
            #     axis.plot(xs, ys, color='black', linewidth=1)

            if select_clusters:
                try:
                    from matplotlib.patches import Ellipse
                except ImportError:
                    raise ImportError('You must have matplotlib.patches available to plot selected clusters.')

                chosen_clusters = self._select_clusters()
                
                # Extract the chosen cluster bounds. If enough duplicate data points exist in the
                # data the lambda value might be infinite. This breaks labeling and highlighting
                # the chosen clusters.
                cluster_bounds = np.array([ plot_data['cluster_bounds'][c] for c in chosen_clusters ])
                if not np.isfinite(cluster_bounds).all():
                    warn('Infinite lambda values encountered in chosen clusters.'
                        ' This might be due to duplicates in the data.')

                # Extract the plot range of the y-axis and set default center and height values for ellipses.
                # Extremly dense clusters might result in near infinite lambda values. Setting max_height
                # based on the percentile should alleviate the impact on plotting.
                plot_range = np.hstack([plot_data['bar_tops'], plot_data['bar_bottoms']])
                plot_range = plot_range[np.isfinite(plot_range)]
                mean_y_center = np.mean([np.max(plot_range), np.min(plot_range)])
                max_height = np.diff(np.percentile(plot_range, q=[10,90]))

                for i, c in enumerate(chosen_clusters):
                    c_bounds = plot_data['cluster_bounds'][c]
                    width = (c_bounds[CB_RIGHT] - c_bounds[CB_LEFT])
                    height = (c_bounds[CB_TOP] - c_bounds[CB_BOTTOM])
                    center = (
                        np.mean([c_bounds[CB_LEFT], c_bounds[CB_RIGHT]]),
                        np.mean([c_bounds[CB_TOP], c_bounds[CB_BOTTOM]]),
                    )
                    
                    # Set center and height to default values if necessary
                    if not np.isfinite(center[1]):
                        center = (center[0], mean_y_center)
                    if not np.isfinite(height):
                        height = max_height

                    # Ensure the ellipse is visible
                    min_height = 0.1*max_height
                    if height < min_height:
                        height = min_height

                    if selection_palette is not None and \
                            len(selection_palette) >= len(chosen_clusters):
                        oval_color = selection_palette[i]
                    else:
                        oval_color = 'r'

                    box = Ellipse(
                        center,
                        2.0 * width,
                        1.2 * height,
                        facecolor='none',
                        edgecolor=oval_color,
                        linewidth=2
                    )

                    if label_clusters:
                        axis.annotate(str(i), xy=center,
                                    xytext=(center[0] - 4.0 * width, center[1] + 0.65 * height),
                                    horizontalalignment='left',
                                    verticalalignment='bottom')

                    axis.add_artist(box)

            if colorbar:
                cb = plt.colorbar(sm, ax=axis)
                if log_size:
                    cb.ax.set_ylabel('log(Number of points)')
                else:
                    cb.ax.set_ylabel('Number of points')

            axis.set_xticks([])
            for side in ('right', 'top', 'bottom'):
                axis.spines[side].set_visible(False)
            axis.invert_yaxis()
            axis.set_ylabel('$\lambda$ value')

            return axis
    
    def _select_clusters(self): 
        """ Computes the clusters based on the stability method

        Parameters
        ----------

        Returns
        """
        if self._clusters_stability is None : 
            self.compute_stablity()

        clusters = _get_selected_clusters(self._raw_tree, self.clusters_stability)
        
        return clusters


    def compute_stablity(self):
        """ Comutes the stability array of the clusters
        """
        self._clusters_stability = _compute_stability(self._raw_tree)
        return self 


    def label_of_stability(self):
        """ Extract the clusters with the stability method
        """
        if self._clusters_stability is None : 
            self.compute_stablity()

        mem_tab, size_tab = _label_of_stability(self._raw_tree, self._clusters_stability)
        clusters = Clustering(mem_tab, size_tab)
        return clusters
    
    def save(self, path):
        """ Save the cluster in a npy file
        
        Parameters
        ----------
            path : path for location where to save the file
        """
        if path[-4:] != ".npy":
            raise AttributeError("The tree is supposed to be stored in a npy file.")
        else : 
            np.save(path, self.raw_tree)   
        return 

